"""Asynchronous Python client for Zinvolt."""

from collections.abc import AsyncGenerator, Generator

import aiohttp
from aioresponses import aioresponses
import pytest

from syrupy import SnapshotAssertion
from zinvolt import ZinvoltClient

from .syrupy import ZinvoltSnapshotExtension


@pytest.fixture(name="snapshot")
def snapshot_assertion(snapshot: SnapshotAssertion) -> SnapshotAssertion:
    """Return snapshot assertion fixture with the Zinvolt extension."""
    return snapshot.use_extension(ZinvoltSnapshotExtension)


@pytest.fixture
async def client() -> AsyncGenerator[ZinvoltClient]:
    """Return a Zinvolt client."""
    async with (
        aiohttp.ClientSession() as session,
        ZinvoltClient("token", session=session) as zinvolt_client,
    ):
        yield zinvolt_client


@pytest.fixture(name="responses")
def aioresponses_fixture() -> Generator[aioresponses]:
    """Return aioresponses fixture."""
    with aioresponses() as mocked_responses:
        yield mocked_responses



"""NetgearPy models for device settings."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import StrEnum

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin


@dataclass
class Battery(DataClassORJSONMixin):
    """Battery model."""

    identifier: str = field(metadata=field_options(alias="id"))
    name: str
    serial_number: str


@dataclass
class BatteryListResponse(DataClassORJSONMixin):
    """Battery list response."""

    batteries: list[Battery]


@dataclass
class CustomMode(DataClassORJSONMixin):
    """Custom mode."""

    mode_id: str = field(metadata=field_options(alias="confId"))
    serial_number: str = field(metadata=field_options(alias="sn"))
    name: str
    weeks: list[int]
    enable: bool


class OnlineStatus(StrEnum):
    """Online status."""

    ONLINE = "ONLINE"
    OFFLINE = "OFFLINE"


class SmartMode(StrEnum):
    """Smart mode."""

    DYNAMIC = "DYNAMIC"
    CHARGED = "CHARGED"
    PERFORMANCE = "PERFORMANCE"
    DEFAULT = "DEFAULT"
    FEED = "FEED"
    CUSTOM = "CUSTOM"
    SELF_USE = "SELF_USE"


@dataclass
class CurrentPower(DataClassORJSONMixin):
    """Current power model.

    All telemetry fields are optional because the API omits them when the
    battery is offline (then only ``onlineStatus`` is returned).
    """

    online_status: OnlineStatus = field(metadata=field_options(alias="onlineStatus"))
    state_of_charge: float | None = field(
        metadata=field_options(alias="soc"), default=None
    )
    output_current: float | None = field(
        metadata=field_options(alias="coc"), default=None
    )
    max_power: int | None = field(metadata=field_options(alias="smp"), default=None)
    on_grid: bool | None = field(metadata=field_options(alias="onGrid"), default=None)
    photovoltaic_power: int | None = field(
        metadata=field_options(alias="ppv"), default=None
    )
    power_socket_output: int | None = field(
        metadata=field_options(alias="pso"), default=None
    )
    is_dormant: bool | None = field(
        metadata=field_options(alias="isDormancy"), default=None
    )


@dataclass
class BatteryState(DataClassORJSONMixin):
    """Battery state."""

    serial_number: str = field(metadata=field_options(alias="sn"))
    name: str
    online_status: OnlineStatus = field(metadata=field_options(alias="onlineStatus"))
    current_power: CurrentPower = field(metadata=field_options(alias="currentPower"))
    smart_mode: SmartMode = field(metadata=field_options(alias="smartMode"))
    global_settings: GlobalSettings = field(
        metadata=field_options(alias="globalSettings")
    )


@dataclass
class OnlineStatusResponse(DataClassORJSONMixin):
    """Online state."""

    serial_number: str = field(metadata=field_options(alias="sn"))
    online_status: OnlineStatus = field(metadata=field_options(alias="onlineStatus"))


@dataclass
class PhotovoltaicData(DataClassORJSONMixin):
    """Photovoltaic data."""

    name: str
    power: int


@dataclass
class GlobalSettings(DataClassORJSONMixin):
    """Global settings."""

    max_output: int = field(metadata=field_options(alias="maxOutput"))
    max_output_limit: int = field(metadata=field_options(alias="maxOutputLimit"))
    max_output_unlocked: bool = field(metadata=field_options(alias="maxOutputUnlocked"))
    battery_upper_threshold: int = field(metadata=field_options(alias="batHighCap"))
    battery_lower_threshold: int = field(metadata=field_options(alias="batUseCap"))
    maximum_charge_power: int = field(metadata=field_options(alias="maxChargePower"))
    standby_time: int = field(metadata=field_options(alias="standbyTime"))


class UnitType(StrEnum):
    """Unit type."""

    INVERTER = "INVERTER"
    EMS = "EMS"
    BATTERY = "BATTERY"


class UnitUpdateStatus(StrEnum):
    """Unit update status."""

    NO_UPDATE = "NO_UPDATE"
    UPDATE_AVAILABLE = "UPDATE_AVAILABLE"
    DISCOVER_UPDATED = "DISCOVER_UPDATED"


@dataclass
class UnitVersion(DataClassORJSONMixin):
    """Unit version information."""

    current_version: str = field(metadata=field_options(alias="currentVersion"))
    status: UnitUpdateStatus
    latest_version: str | None = field(
        metadata=field_options(alias="latestVersion"), default=None
    )


@dataclass
class Unit(DataClassORJSONMixin):
    """Unit model."""

    serial_number: str = field(metadata=field_options(alias="usn"))
    name: str
    type: UnitType
    abnormal_amount: int = field(metadata=field_options(alias="abnormalAmount"))
    resettable: bool
    version: UnitVersion


@dataclass
class UnitsResponse(DataClassORJSONMixin):
    """Units list response."""

    units: list[Unit]


@dataclass
class BatteryUnit(DataClassORJSONMixin):
    """Battery unit."""

    serial_number: str = field(metadata=field_options(alias="usn"))
    version: Version
    points: list[Checkpoint]
    battery_model: str = field(metadata=field_options(alias="batteryModel"))
    power: int


@dataclass
class Version(DataClassORJSONMixin):
    """Version information."""

    current_version: str = field(metadata=field_options(alias="currentVersion"))
    latest_version: str | None = field(
        metadata=field_options(alias="latestVersion"), default=None
    )


@dataclass
class Checkpoint(DataClassORJSONMixin):
    """Checkpoint information."""

    point: str
    normal: bool



"""Asynchronous Python client for Zinvolt."""

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from http import HTTPStatus
import logging
from typing import Any, Self

from aiohttp import ClientError, ClientSession
from aiohttp.hdrs import METH_GET, METH_POST, METH_PUT
from mashumaro.codecs.orjson import ORJSONDecoder
import orjson
from yarl import URL

from zinvolt.exceptions import ZinvoltAuthenticationError, ZinvoltError
from zinvolt.models import (
    Battery,
    BatteryListResponse,
    BatteryState,
    BatteryUnit,
    CustomMode,
    GlobalSettings,
    OnlineStatus,
    OnlineStatusResponse,
    PhotovoltaicData,
    SmartMode,
    Unit,
    UnitsResponse,
)

VERSION = "1"

_LOGGER = logging.getLogger(__package__)

HOST = "app.zinvolt.com"


@dataclass
class ZinvoltClient:
    """Main class for handling connections with the Zinvolt API."""

    token: str | None = None
    session: ClientSession | None = None
    request_timeout: int = 10
    _close_session: bool = False

    async def _request(
        self,
        uri: str,
        *,
        method: str = METH_GET,
        data: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
    ) -> str:
        """Handle a request to the Zinvolt API."""
        url = URL.build(host=HOST, scheme="https").joinpath(f"api/public/v2/{uri}")
        base_headers = {
            "User-Agent": f"python-zinvolt/{VERSION}",
        }
        if self.token:
            base_headers["Authorization"] = f"Bearer {self.token}"

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        if headers is None:
            headers = {}

        try:
            async with asyncio.timeout(self.request_timeout):
                response = await self.session.request(
                    method,
                    url,
                    headers=base_headers | headers,
                    json=data,
                )
                body = await response.text()
        except TimeoutError as err:
            msg = f"Timeout occurred while connecting to {url}"
            raise ZinvoltError(msg) from err
        except ClientError as err:
            msg = f"Error occurred while communicating with {url}: {err}"
            raise ZinvoltError(msg) from err

        if response.status in (HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN):
            msg = f"Authentication failed for {url} (status {response.status})"
            raise ZinvoltAuthenticationError(msg)
        if response.status >= HTTPStatus.BAD_REQUEST:
            # The API returns JSON error envelopes like
            # {"message": "...", "type": "...", "code": 400}; surface the
            # message when it's there so callers don't have to parse it.
            detail = body
            try:
                payload = orjson.loads(body)  # pylint: disable=no-member
            except orjson.JSONDecodeError:  # pylint: disable=no-member
                pass
            else:
                if isinstance(payload, dict) and "message" in payload:
                    detail = str(payload["message"])
            msg = f"Unexpected status {response.status} from {url}: {detail}"
            raise ZinvoltError(msg)

        return body

    async def _get(self, uri: str) -> str:
        """Handle a GET request to the Zinvolt API."""
        return await self._request(uri, method=METH_GET)

    async def _put(self, uri: str, data: dict[str, Any]) -> str:
        """Handle a PUT request to the Zinvolt API."""
        return await self._request(uri, method=METH_PUT, data=data)

    async def _post(self, uri: str, data: dict[str, Any]) -> str:
        """Handle a POST request to the Zinvolt API."""
        return await self._request(uri, method=METH_POST, data=data)

    async def login(self, email: str, password: str) -> str:
        """Login to the Zinvolt API."""
        result = await self._post("login", {"email": email, "password": password})
        self.token = orjson.loads(result)["token"]  # pylint: disable=no-member
        return self.token

    async def get_batteries(self) -> list[Battery]:
        """Retrieve the list of batteries from the Zinvolt API."""
        result = await self._get("system/batteries")
        return BatteryListResponse.from_json(result).batteries

    async def get_battery_status(self, battery_id: str) -> BatteryState:
        """Retrieve the battery status for the given battery ID."""
        result = await self._get(f"system/{battery_id}/basic/current-state")
        return BatteryState.from_json(result)

    async def is_battery_online(self, battery_id: str) -> bool:
        """Retrieve the battery status for the given battery ID."""
        result = await self._get(f"system/{battery_id}/basic/online-status")
        return (
            OnlineStatusResponse.from_json(result).online_status is OnlineStatus.ONLINE
        )

    async def get_photovoltaic_data(self, battery_id: str) -> list[PhotovoltaicData]:
        """Retrieve the photovoltaic data for the given battery ID."""
        result = await self._get(f"system/{battery_id}/basic/pv-data")
        return ORJSONDecoder(list[PhotovoltaicData]).decode(result)

    async def get_global_settings(self, battery_id: str) -> GlobalSettings:
        """Retrieve the global settings for the given battery ID."""
        result = await self._get(f"system/{battery_id}/configuration/global-settings")
        return GlobalSettings.from_json(result)

    async def get_custom_modes(self, battery_id: str) -> list[CustomMode]:
        """Retrieve the custom modes for the given battery ID."""
        result = await self._get(f"system/{battery_id}/custom-mode")
        return ORJSONDecoder(list[CustomMode]).decode(result)

    async def get_units(self, battery_id: str) -> list[Unit]:
        """Retrieve all units for the given battery ID."""
        result = await self._get(f"system/{battery_id}/unit")
        return UnitsResponse.from_json(result).units

    async def get_battery_unit(
        self, battery_id: str, battery_serial_number: str
    ) -> BatteryUnit:
        """Retrieve the battery unit for the given battery ID."""
        result = await self._get(
            f"system/{battery_id}/unit/battery/{battery_serial_number}"
        )
        return BatteryUnit.from_json(result)

    async def set_smart_mode(
        self, battery_id: str, smart_mode: SmartMode | str
    ) -> None:
        """Retrieve the custom modes for the given battery ID."""
        data: dict[str, Any] = {}
        if not isinstance(smart_mode, SmartMode):
            data["custom_mode_id"] = smart_mode
            data["mode"] = SmartMode.CUSTOM
        else:
            data["mode"] = smart_mode
        await self._put(f"system/{battery_id}/operation/switch-smart-mode", data=data)

    async def set_max_output(self, battery_id: str, max_output: int) -> None:
        """Set the maximum output for the given battery ID."""
        await self._post(
            f"system/{battery_id}/configuration/global-settings",
            data={
                "max_output": max_output,
            },
        )

    async def set_lower_threshold(self, battery_id: str, lower_threshold: int) -> None:
        """Set the lower threshold for the given battery ID."""
        await self._post(
            f"system/{battery_id}/configuration/global-settings",
            data={
                "bat_use_cap": lower_threshold,
            },
        )

    async def set_upper_threshold(self, battery_id: str, upper_threshold: int) -> None:
        """Set the upper threshold for the given battery ID."""
        await self._post(
            f"system/{battery_id}/configuration/global-settings",
            data={
                "max_charge_power": upper_threshold,
            },
        )

    async def set_standby_time(self, battery_id: str, standby_time: int) -> None:
        """Set the standby time in minutes for the given battery ID."""
        await self._post(
            f"system/{battery_id}/configuration/global-settings",
            data={
                "standby_time": standby_time,
            },
        )

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The ZinvoltClient object.

        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.

        """
        await self.close()



"""Asynchronous Python client for Netgear routers."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from aiohttp.hdrs import METH_GET, METH_POST, METH_PUT
from aioresponses import aioresponses
import pytest

from zinvolt.exceptions import ZinvoltAuthenticationError, ZinvoltError
from zinvolt.models import SmartMode

from . import load_fixture
from .const import HEADERS, URL

if TYPE_CHECKING:
    from syrupy import SnapshotAssertion
    from zinvolt.zinvolt import ZinvoltClient


async def test_login(
    responses: aioresponses,
    client: ZinvoltClient,
    snapshot: SnapshotAssertion,
) -> None:
    """Test loggign in."""
    responses.post(
        f"{URL}login",
        status=200,
        body=load_fixture("login.json"),
    )
    assert await client.login(email="test@test.com", password="abc") == snapshot
    responses.assert_called_once_with(
        f"{URL}login",
        METH_POST,
        headers=HEADERS,
        json={
            "email": "test@test.com",
            "password": "abc",
        },
    )


async def test_set_smart_mode(responses: aioresponses, client: ZinvoltClient) -> None:
    """Test setting smart mode."""
    responses.put(
        f"{URL}system/123123/operation/switch-smart-mode",
        status=200,
    )
    await client.set_smart_mode(battery_id="123123", smart_mode=SmartMode.DYNAMIC)
    responses.assert_called_once_with(
        f"{URL}system/123123/operation/switch-smart-mode",
        METH_PUT,
        headers=HEADERS,
        json={"mode": "DYNAMIC"},
    )


async def test_set_custom_smart_mode(
    responses: aioresponses, client: ZinvoltClient
) -> None:
    """Test setting smart mode."""
    responses.put(
        f"{URL}system/123123/operation/switch-smart-mode",
        status=200,
    )
    await client.set_smart_mode(
        battery_id="123123", smart_mode="EdxM2EsVNJVc6abhbVnbeby5bcq5EBXh"
    )
    responses.assert_called_once_with(
        f"{URL}system/123123/operation/switch-smart-mode",
        METH_PUT,
        headers=HEADERS,
        json={
            "mode": "CUSTOM",
            "custom_mode_id": "EdxM2EsVNJVc6abhbVnbeby5bcq5EBXh",
        },
    )


@pytest.mark.parametrize(
    ("method", "kwargs", "endpoint", "fixture"),
    [
        ("get_batteries", {}, "system/batteries", "batteries"),
        (
            "get_battery_status",
            {"battery_id": "123123"},
            "system/123123/basic/current-state",
            "current_state",
        ),
        (
            "is_battery_online",
            {"battery_id": "123123"},
            "system/123123/basic/online-status",
            "online_status",
        ),
        (
            "get_photovoltaic_data",
            {"battery_id": "123123"},
            "system/123123/basic/pv-data",
            "pv_data",
        ),
        (
            "get_global_settings",
            {"battery_id": "123123"},
            "system/123123/configuration/global-settings",
            "global_settings",
        ),
        (
            "get_custom_modes",
            {"battery_id": "123123"},
            "system/123123/custom-mode",
            "custom_modes",
        ),
        (
            "get_battery_unit",
            {"battery_id": "123123", "battery_serial_number": "abc123"},
            "system/123123/unit/battery/abc123",
            "battery_unit",
        ),
        (
            "get_units",
            {"battery_id": "123123"},
            "system/123123/unit",
            "unit",
        ),
    ],
    ids=[
        "get_batteries",
        "get_battery_status",
        "is_battery_online",
        "get_photovoltaic_data",
        "get_global_settings",
        "get_custom_modes",
        "get_battery_unit",
        "get_units",
    ],
)
async def test_retrieve_data(
    responses: aioresponses,
    client: ZinvoltClient,
    snapshot: SnapshotAssertion,
    method: str,
    kwargs: dict[str, Any],
    endpoint: str,
    fixture: str,
) -> None:
    """Test retrieving data."""
    responses.get(
        f"{URL}{endpoint}",
        status=200,
        body=load_fixture(f"{fixture}.json"),
    )
    assert await getattr(client, method)(**kwargs) == snapshot
    responses.assert_called_once_with(
        f"{URL}{endpoint}",
        METH_GET,
        headers=HEADERS,
        json=None,
    )


@pytest.mark.parametrize(
    ("method", "kwargs", "data"),
    [
        (
            "set_max_output",
            {"max_output": 500},
            {"max_output": 500},
        ),
        (
            "set_lower_threshold",
            {"lower_threshold": 500},
            {"bat_use_cap": 500},
        ),
        (
            "set_upper_threshold",
            {"upper_threshold": 500},
            {"max_charge_power": 500},
        ),
        (
            "set_standby_time",
            {"standby_time": 500},
            {"standby_time": 500},
        ),
    ],
    ids=[
        "set_max_output",
        "set_lower_threshold",
        "set_upper_threshold",
        "set_standby_time",
    ],
)
async def test_setting_global_settings(
    responses: aioresponses,
    client: ZinvoltClient,
    method: str,
    kwargs: dict[str, Any],
    data: dict[str, Any],
) -> None:
    """Test setting global settings."""
    responses.post(
        f"{URL}system/123123/configuration/global-settings",
        status=200,
    )
    await getattr(client, method)(**(kwargs | {"battery_id": "123123"}))
    responses.assert_called_once_with(
        f"{URL}system/123123/configuration/global-settings",
        METH_POST,
        headers=HEADERS,
        json=data,
    )


async def test_get_battery_status_offline(
    responses: aioresponses,
    client: ZinvoltClient,
    snapshot: SnapshotAssertion,
) -> None:
    """Offline batteries return currentPower with only onlineStatus."""
    responses.get(
        f"{URL}system/123123/basic/current-state",
        status=200,
        body=load_fixture("current_state_offline.json"),
    )
    assert await client.get_battery_status(battery_id="123123") == snapshot


@pytest.mark.parametrize(
    ("status", "body", "expected"),
    [
        (400, load_fixture("device_not_exist.json"), "The device does not exist"),
        (500, "internal server error", "internal server error"),
        (404, "", ""),
    ],
    ids=["json_envelope", "plain_text", "empty_body"],
)
async def test_request_error_status(
    responses: aioresponses,
    client: ZinvoltClient,
    status: int,
    body: str,
    expected: str,
) -> None:
    """Non-2xx responses raise ZinvoltError with the API message when present."""
    responses.get(
        f"{URL}system/123123/unit/battery/abc",
        status=status,
        body=body,
    )
    with pytest.raises(ZinvoltError) as err:
        await client.get_battery_unit(battery_id="123123", battery_serial_number="abc")
    assert expected in str(err.value)


@pytest.mark.parametrize("status", [401, 403])
async def test_request_authentication_error(
    responses: aioresponses,
    client: ZinvoltClient,
    status: int,
) -> None:
    """401/403 responses raise ZinvoltAuthenticationError."""
    responses.get(
        f"{URL}system/batteries",
        status=status,
        body='{"message":"Unauthenticated."}',
    )
    with pytest.raises(ZinvoltAuthenticationError):
        await client.get_batteries()
